//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: DHCPWatcher
//     Author: Jim Guggemos
//     Description: Class that monitors DHCP packets from PacketQueue
//
// $Date$
//
//***************************************************************************

#include "DHCPWatcher.h"
#include "HashManager.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include "SAPQ.h"
#include <string.h>

#define SYNCHRONIZE_DISPATCHER() Mutex::Holder _dmLock(this->m_dispMutex)

DHCPWatcher::DHCPWatcher(HashManager &hashmgr, int queueNum, bool reverse) :
    m_dispatcher(0), m_threadId(0), m_hashmgr(hashmgr), m_queueNum(queueNum), m_reverse(reverse),
    m_dispMutex("DHCPWatcher::m_dispMutex")
{
}

DHCPWatcher::~DHCPWatcher()
{
    stop();
}

bool
DHCPWatcher::start()
{
    try
    {
        {
            SYNCHRONIZE_DISPATCHER();
            m_dispatcher = new PacketDispatcher( 2048 );

            // This gets deleted automatically when the dispatcher exits
            PacketQueue *pq = new PacketQueue( m_queueNum, this );
            m_dispatcher->addQueue( pq );
            logDebug( "DHCPWatcher PacketQueue created, bound to queue #%d",
                      m_queueNum );
        }
    }
    catch (PacketQueueError &e)
    {
        logMessage( "PacketQueue Exception: %s", e.what() );
        return false;
    }

    ServiceMonitor::instance()->setDHCPWatcher( true );

    pthread_attr_t attr;
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
    if ( pthread_create( &m_threadId, &attr, threadRun, (void*)this ) != 0 )
    {
        ServiceMonitor::instance()->setDHCPWatcher( false );
        return false;
    }
    pthread_attr_destroy( &attr );		// No effect in LINUX

    return true;
}

void
DHCPWatcher::stop()
{
    { // protects us from attempting to stop a null m_dispatcher
        SYNCHRONIZE_DISPATCHER();
        if ( m_dispatcher )
            m_dispatcher->stop();
    }

    logMessage( "DHCPWatcher stopped" );

    ServiceMonitor::instance()->setDHCPWatcher( false );
}

void
DHCPWatcher::run()
{
    logMessage( "DHCPWatcher started" );

    // Obviously, don't sync on dispatcher here because stop could never
    // run.  m_dispatcher shuold not be null here.
    m_dispatcher->run();

    {
        SYNCHRONIZE_DISPATCHER();
        delete m_dispatcher;
        m_dispatcher = 0;
    }

    ServiceMonitor::instance()->setDHCPWatcher( false );

}

bool
DHCPWatcher::onPacket(unsigned char **payload, int *payloadLen)
{
    HwAddr chaddr;
    bool goodPkt;
    bool accept;
    in_addr_t srcaddr;
    in_addr_t giaddr;

    if ( (goodPkt=decodePacket( *payload, *payloadLen, chaddr, srcaddr, giaddr )) )
    {
        if (m_hashmgr.lookup( chaddr ))
        {
            const char *tag = chaddr.getTag();

            if (!m_reverse)
            {
                logDebug( "Packet received from %s: Found and ALLOWED", chaddr.ascii().c_str() );
                if( tag && strlen(tag) > 0 )
                {
                    logDebug( "Writing tag '%s' for %s", tag, chaddr.ascii().c_str() );
                    writeTag( *payload, *payloadLen, tag );
                }
                else
                {
                    logDebug( "No tag for %s", chaddr.ascii().c_str() );
                }

                accept = true;
            }
            else
            {
                logDebug( "Packet received from %s: Found and DENIED", chaddr.ascii().c_str() );
                accept = false;
            }
        }
        else
        {
            if (!m_reverse)
            {
                logDebug( "Packet received from %s: Not found and DENIED", chaddr.ascii().c_str() );
                accept = false;
            }
            else
            {
                logDebug( "Packet received from %s: Not found and ALLOWED", chaddr.ascii().c_str() );
                accept = true;
            }
        }
    }
    else
        accept = false;

    if ( ! goodPkt )
        logMessage( "Improperly formatted DHCP/BOOTP packet received");

    return accept;
}

void
DHCPWatcher::writeTag(unsigned char *data, int len, const char *tag)
{
    // Size has already been verified by decodePacket, so we know the
    // buffer is big enough, but we still need to extract the ip header len
    // to get the right offset into the packet.

    int ipHdrLen = (data[0] & 0xf ) * 4;

    // Zero out UDP checksum instead of recalculating
    udphdr *hdr = (udphdr*)(data + ipHdrLen);
    hdr->cksum = 0;
       
    // Write the tag to the BOOTP server host name field + 1
    // The preceding null byte allows the DHCP server to correctly
    // parse the packet if there is an override option per
    // RFC2132(9.3)
    strncpy( (char*)&data[ipHdrLen+8+45], tag, 32 );
    data[ipHdrLen+8+44]= '\0';
}

bool
DHCPWatcher::decodePacket(const unsigned char *data, const int len,
                          HwAddr &ha, in_addr_t &srcip, in_addr_t &giaddr)
{
    // The packet payload starts with the IP header; the ethernet frame
    // has already been stripped off.

    // Check to see if we have enough bytes for an minimal IP header.
    if ( len < 20 )
        return false;

    // Check version nybble for IPv4
    if ( (data[0] >> 4) & 0xf != 4 )
        return false;

    int ipHdrLen = (data[0] & 0xf ) * 4;

    // Must be at least 20 bytes
    if ( ipHdrLen < 20 )
        return false;

    // Now check if we have enough bytes for the minimal DHCP packet with UDP
    // header.
    else if ( len < ipHdrLen + 8/*udp*/ + 228/*min bootp/dhcp*/ )
        return false;

    //Extract source address from IP header
    in_addr addr;

    memcpy((void*)&addr, &data[12], 4 );
    srcip = ntohl(addr.s_addr);

    // Check IP header for packet protocol for UDP packet (proto=17)
    if ( data[9] != 17 )
        return false;

    {
        unsigned short destPort;
        memcpy( (void*)&destPort, &data[ipHdrLen+2], 2 );
        destPort = ntohs( destPort );

        // Check destination for BOOTP/DHCP server
        if ( destPort != 67 )
            return false;
    }

    // Extract GIADDR (relay agent IP address) at offset 24 into
    // BOOTP header; UDP header is 8 bytes.
    memcpy((void*)&addr.s_addr, &data[ipHdrLen+8+24], 4 );
    giaddr = ntohl(addr.s_addr);

    // The chaddr (client hardware address) is at offset 28 into the
    // BOOTP header

    try
    {
        ha = HwAddr( &data[ipHdrLen+8+28] );        
    }
    catch ( std::exception &e )
    {
        // Invalid ethernet address
        return false;
    }

    return true;
}

void *
DHCPWatcher::threadRun(void *vthis)
{
    DHCPWatcher *rthis = reinterpret_cast<DHCPWatcher *>(vthis);

    rthis->run();
    return NULL;
}
