//***************************************************************************
//
// $Source$
// $Revision: 7388 $
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2007 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: HwAddrHash
//     Author: Mark Manes
//     Description: Multi-table wrapper around HwAddrHash.
//
//
//***************************************************************************

#include "HashManager.h"

HashManager::HashManager(int minBuckets)
{
    for(int i = 0; i < 6; i++)
        hashes[i] = new HwAddrHash(minBuckets);
}

HashManager::~HashManager()
{
    for(int i = 0; i < 6; i++)
        delete hashes[i];
}

HwAddrHash* HashManager::getHash(int size) throw(std::invalid_argument)
{    
    if(size < 1 || size > 6 )
        throw std::invalid_argument("HwAddr has invalid size");
    
    return hashes[size - 1];
}

HwAddrHash* HashManager::getHash(const HwAddr &hw) throw(std::invalid_argument)
{
    return getHash(hw.size());
}

void HashManager::put(const HwAddr &hw) throw(std::invalid_argument)
{
    getHash(hw)->put(hw);
}


bool HashManager::remove(const HwAddr &hw) throw(std::invalid_argument)
{
    return getHash(hw)->remove(hw);
}

/**
 * Only looks in matching sized table.
 */
bool HashManager::exists(const HwAddr &hw) throw(std::invalid_argument)
{
    return getHash(hw)->exists(hw);
}

/**
 * Returns true if hw exists in any of the tables, overwrites hw with copy of key in map.
 */
bool HashManager::lookup(HwAddr &hw) const throw(std::invalid_argument)
{
    for(int i = 0; i < hw.size(); i++)
    {
        if(!hashes[i]->empty())
        {
            int dummy;
            HwAddr smallHw(hw, i + 1);
            if(hashes[i]->get(smallHw, dummy))
            {
                hw = smallHw;
                return true;
            }
        }
    }
    
    return false;
}

bool HashManager::getAll(HwAddrHash::AddrList &alist, int size/*=-1*/) const
{
    if(size == -1)
    {
        for(int i = 0; i < 6; i++)
        {
            if(!hashes[i]->empty())
            {
                hashes[i]->getAll(alist);
            }
        }
    }
    else if( size > 0 && size <= 6)
    {
        hashes[size - 1]->getAll(alist);
    }    
    
    return (! alist.empty());
}


void HashManager::clear()
{
    for(int i = 0; i < 6; i++)
    {
        if(!hashes[i]->empty())    
            hashes[i]->clear();
    }
}

/**
 * Returns a string containing the stats of all tables if size == -1 or is omitted, 
 * or the table for specified MAC prefix size.
 */
string HashManager::generateStats(int size/*=-1*/) const
{
    char x[100];
    
    string s;
    
    if(size == -1)
    {
        for(int i = 0; i < 6; i++)
        {
            sprintf(x, "*** MAC Hash Statistics, MAC size: %d ***\n", i + 1);
            s += x;
            s += hashes[i]->generateStats();
        }            
    }
    else if( size > 0 && size <= 6)
    {
        sprintf(x, "*** MAC Hash Statistics, MAC size: %d ***\n", size);
        s += x;
        s += hashes[size - 1]->generateStats();
    }
    else
    {
        s = sprintf(x, "Invalid size: %d", size);
    }
    
    return s;
}

// ---- Unit test definition -------------------

#ifdef UNITTEST
void HashManager::test() throw(std::logic_error)
{
    int i;    
    string macs[6];
    HwAddr hw;
    macs[0] = string("02");  
    macs[1] = string("00:22");
    macs[2] = string("00:33:33");
    macs[3] = string("00:44:44:44");
    macs[4] = string("00:55:55:55:55");
    macs[5] = string("00:66:66:66:66:66");    
    
    HashManager mgr(1571);
        
    for(i = 0; i < 6; i++)
    {
        hw = HwAddr(macs[i]);
        mgr.put(hw);
    }
    
    for(i = 0; i < 6; i++)
    {
        hw = HwAddr(macs[i]);
        if(!mgr.getHash(i + 1)->exists(hw))
            throw std::logic_error("Address didn't get stored in map");
        
        if(mgr.getHash(i+1)->count() != 1)
            throw std::logic_error("Hash has wrong size");
        
        if(!mgr.lookup(hw))
            throw std::logic_error("Address failed lookup on itself");
    }
    
    hw = HwAddr("02:AA");
    if(!mgr.lookup(hw)) throw std::logic_error("Address failed prefix lookup");
    hw = HwAddr("00:22:AA");
    if(!mgr.lookup(hw)) throw std::logic_error("Address failed prefix lookup");
    hw = HwAddr("02:33:33:DD");
    if(!mgr.lookup(hw)) throw std::logic_error("Address failed prefix lookup");
    hw = HwAddr("02:44:44:44:99:FF");
    if(!mgr.lookup(hw)) throw std::logic_error("Address failed prefix lookup");    
    hw = HwAddr("02:55:55:55:55:AA");
    if(!mgr.lookup(hw)) throw std::logic_error("Address failed prefix lookup");    
    
    
    
    // This may take some time....    
    unsigned char base[3] = {0x00, 0x11, 0x12};
    unsigned char mac[6]  = {0x00, 0x11, 0x12};
    
    mgr.clear();    
    hw = HwAddr(base, 3);
    mgr.put(hw);
    
    for(unsigned char x = 0; x < 255; x++)
    {
        for(unsigned char  y = 0; y < 255; y++)
        {
            for(unsigned char  z = 0; z < 255; z++)
            {
                mac[3] = x;
                mac[4] = y;
                mac[5] = z;
                hw = HwAddr(mac);
                if(!mgr.lookup(hw))
                    throw std::logic_error("Couldn't find MAC by prefix");
            }
        }
    }
    
    mac[0] = 0x22;
    for(unsigned char x = 0; x < 255; x++)
    {
        for(unsigned char  y = 0; y < 255; y++)
        {
            for(unsigned char  z = 0; z < 255; z++)
            {
                mac[3] = x;
                mac[4] = y;
                mac[5] = z;
                hw = HwAddr(mac);
                if(mgr.lookup(hw))
                    throw std::logic_error("Should not have found MAC by prefix");
            }
        }
    }
    
}
#endif // UNITTEST
