//***************************************************************************
//
// $Source$
// $Revision: 70266 $
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: HwAddrHash
//     Author: Jim Guggemos
//     Description: Classes to maintain thread-safe hash table of allowable
//                  hardware addresses.
//
// $Date: 2010-06-16 15:53:55 -0600 (Wed, 16 Jun 2010) $
//
//***************************************************************************

#ifndef __HwAddrHash_H__
#define __HwAddrHash_H__

#include <ext/hash_map>
#include <vector>
#include "Mutex.h"
#include <stdexcept>
#include <stdio.h>


class HwAddr
{
public:
    HwAddr() throw();
    HwAddr(const unsigned char *buf, int len=6) throw(std::invalid_argument);

    HwAddr(const string &ascii) throw(std::invalid_argument);
    
    HwAddr(const HwAddr &hw, int size) throw(std::invalid_argument);
    
    HwAddr &operator=(const HwAddr &obj) throw(std::invalid_argument);

    unsigned char operator[](const int i) const throw(std::out_of_range);

    size_t hashkey() const throw();

    bool operator==(const HwAddr &hw) const throw();

    string ascii() const throw();

    // Returns the significant digits of this HW addr
    int size() const;

    // Sets the value to be written to the outgoing packet's BOOTP filename field
    void setTag(const char* tag);

    // Returns the value to the tag
    const char* getTag() const;

#ifdef UNITTEST
    static void test() throw();
#endif
private:
    unsigned char bytes[6];
    char tag[32];
    int addrsize;
    void checkValidity() const throw(std::invalid_argument);
    static unsigned char hexval(const char c) throw();
    void cpyMem(const unsigned char *buf) throw(std::invalid_argument);
};

class HwAddrHash
{
public:
    HwAddrHash(int minBuckets);
    ~HwAddrHash();

    // Put key/value pair into hash; value isn't used at this time.
    void put(const HwAddr &hw, int val=1);

    // Remove HwAddr from hash; returns bool if HwAddr was present.
    bool remove(const HwAddr &hw);

    // Returns true if hw exists and returns value in val
    bool get(HwAddr &hw, int &val) const;

    // Returns true if hw exists
    bool exists(const HwAddr &hw) const;

    typedef vector<HwAddr> AddrList;
    bool getAll(HwAddrHash::AddrList &alist) const;

    // Clear ALL values in hash
    void clear();

    // Generate a string containing statistics about the hash
    string generateStats() const;

    // Return count of number of items in hash.
    int count() const { return m_hash.size(); }

    // Returns true if empty.
    bool empty() const { return m_hash.empty(); }


   class HashFunc
    {
    public:
        size_t operator()(const HwAddr &hw) const;
    };

    class EqualKey
    {
    public:
        bool operator()(const HwAddr &hw1, const HwAddr &hw2) const;
    };


private:

    typedef __gnu_cxx::hash_map<HwAddr, int, HashFunc, EqualKey> HashType;

    mutable Mutex m_mutex;
    HashType m_hash;
};

#endif // __HwAddrHash_H__
