//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: Mutex
//     Author: Jim Guggemos
//     Description: Definition for the Mutex and associated class
//
// $Date$
//
//***************************************************************************

#include "Mutex.h"
#include <sys/time.h>

/// Construct an unnamed Mutex.
Mutex::Mutex(MutexType type)
{
    init( type );
}

/// Construct a named Mutex.
Mutex::Mutex(const string &name, MutexType type) : m_name( name )
{
    init( type );
}

/// Initialize the object
void
Mutex::init(MutexType type)
{
    pthread_mutexattr_t attr;
    int kind;

    pthread_mutexattr_init( &attr );

    switch ( type )
    {
	case SAFE:
	    kind = PTHREAD_MUTEX_ERRORCHECK_NP;
	    break;

	case RECURSIVE:
	    kind = PTHREAD_MUTEX_RECURSIVE_NP;
	    break;

	case FAST:
	default:
	    kind = PTHREAD_MUTEX_FAST_NP;
	    break;
    }

    pthread_mutexattr_settype( &attr, kind );
    
    pthread_mutex_init( &m_mutex, &attr );

    pthread_mutexattr_destroy( &attr );		/* NO-OP in linux */
}


/// Destructor
Mutex::~Mutex()
{
    pthread_mutex_destroy( &m_mutex );		/* NO-OP in linux */
}


Mutex::Holder::Holder(Mutex &mutex) : m_mutex( mutex )
{
    m_mutex.lock();
}

Mutex::Holder::Holder(ConditionVar &condx) : m_mutex( condx.mutex() )
{
    m_mutex.lock();
}

Mutex::Holder::~Holder()
{
    m_mutex.unlock();
}




/// Construct an unnamed ConditionVar with a locally constructed Mutex
ConditionVar::ConditionVar() :
	m_mutex( new Mutex(Mutex::FAST) ),
	m_ourMutex( true )
{
    pthread_cond_init( &m_cond, NULL );
}

/// Construct a named ConditionVar with a locally constructed Mutex
ConditionVar::ConditionVar(const string &name) :
	m_mutex( new Mutex(name, Mutex::FAST) ),
	m_ourMutex( true ),
	m_name( name )
{
    pthread_cond_init( &m_cond, NULL );
}

/// Construct an unnamed ConditionVar with given Mutex
ConditionVar::ConditionVar(Mutex *mutex) :
	m_mutex( mutex ),
	m_ourMutex( false )
{
    pthread_cond_init( &m_cond, NULL );
}

/// Construct a named ConditionVar with given Mutex
ConditionVar::ConditionVar(const string &name, Mutex *mutex) :
	m_mutex( mutex ),
	m_ourMutex( false ),
	m_name( name )
{
    pthread_cond_init( &m_cond, NULL );
}

/// Destructor
ConditionVar::~ConditionVar()
{
    pthread_cond_destroy( &m_cond );		/* NO-OP in linux */

    if ( m_ourMutex && m_mutex )
	delete m_mutex;
}

/** Wait (block) on the ConditionVar.  If a timeout is specified, the
    thread will wake up after timeout seconds if no other thread signaled
    the condition.  Returns true if condition was signaled, false if a 
    timeout occurred.
    NOTE THAT MUTEX MUST BE LOCKED BEFORE CALLING THIS METHOD.
*/
bool
ConditionVar::wait(int timeout)
{
    bool retval = true;

    if ( timeout > 0 )
    {
	struct timeval now;
	struct timespec then;
	int rval = -1;

	gettimeofday( &now, NULL );
	then.tv_sec = now.tv_sec + timeout;
	then.tv_nsec = now.tv_usec * 1000;

	while ( rval != 0 && rval != ETIMEDOUT )
	    rval = pthread_cond_timedwait( &m_cond, &m_mutex->m_mutex,
					   &then );

	retval = (rval != ETIMEDOUT);
    }
    
    else
	pthread_cond_wait( &m_cond, &m_mutex->m_mutex );

    return retval;
}



/** Allocate a Semaphore with the given initial_value.  Use
    1 for initial_value if you want mutex-like behavior. */
Semaphore::Semaphore(int initial_value)
{
    init( initial_value );
}

/// Allocate a named semaphore
Semaphore::Semaphore(const string &name, int initial_value) : m_name(name)
{
    init( initial_value );
}

/// Destructor
Semaphore::~Semaphore()
{
    sem_destroy( &m_semaphore );		/* NO-OP in linux */
}

void
Semaphore::init(int initial_value)
{
    sem_init( &m_semaphore, 0, initial_value );
}

/** Blocks thread until semaphore has non-zero count; then atomically
    decrements the counter. Returns false if some error other than
    EINTR occurred (will re-wait on EINTR).  If false is returned,
    the "wait" operation did not succeed. */
bool
Semaphore::wait()
{
    bool success = true;
    int rval;

    // Retry if we were interrupted by a signal.
    while ( (rval = sem_wait( &m_semaphore )) < 0 && errno == EINTR );

    if ( rval < 0 )
    {
	perror( "Semaphore::wait sem_wait" );
	success = false;
    }

    return success;
}
