//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: PacketQueue
//     Author: Jim Guggemos
//     Description: Class to interface with libnetfilter_queue
//
// $Date$
//
//***************************************************************************

#ifndef __PacketQueue_H__
#define __PacketQueue_H__

#include <string>
#include "Mutex.h"
#include <set>

extern "C" {
    #include <netinet/in.h>
    #include <linux/netfilter.h>
    #include <libnetfilter_queue/libnetfilter_queue.h>
}

/// Common exception that may be thrown by these classes
class PacketQueueError : public std::exception
{
public:
    PacketQueueError(const std::string &method, const std::string &message);
    ~PacketQueueError() throw() {}

    std::string getText() const { return m_text; }
    virtual const char *what() const throw() { return getText().c_str(); }

    
private:
    friend class PacketQueue;

    std::string m_text;
};

/**
 * "Interface" required to create a PacketQueue.
 */
class PacketQueueListener
{
public:
    virtual ~PacketQueueListener() {}

    /**
     * Called when a packet has arrived on the queue.
     * If this handler throws an exception, the PacketQueue
     * will shut down.
     *
     * @param payload The actual packet data
     * @param payloadLen Length of payload buffer
     * @return true to accept packet; false to drop it.
     */
    virtual bool onPacket(unsigned char **payload, int *payloadLen) = 0;
};

// Fwd declaration
class PacketDispatcher;

class PacketQueue
{
public:
    /**
     * Constructs a PacketQueue to listen on a given queueNum.  This
     * object must be handed to a PacketDispatcher via addQueue().
     * Packets will only begin to arrive at the queue and call back
     * the PacketQueueListener once PacketQueue::run() has been started.
     *
     * @param queueNum Indicates what NFQ queue number should be listened to
     * @param listener Object called back when a packet is received
     * @throws PacketQueueError if there was some problem with the args
     */
    PacketQueue(int queueNum, PacketQueueListener *listener)
        throw(PacketQueueError);

    /**
     * Destructor - will close the queue.
     */
    virtual ~PacketQueue() throw();

private:
    friend class PacketDispatcher;

    void createQueueHandle(PacketDispatcher *dispatcher, int bufsiz)
        throw(PacketQueueError); 

    bool callListener(unsigned char** payload, int *len) throw();

    static int callback_stub(struct nfq_q_handle *, struct nfgenmsg *,
                             struct nfq_data *, void *);

    PacketQueueListener *const m_listener;
    const int m_queueNum;
    struct nfq_q_handle *m_queueHandle;
    PacketDispatcher *m_dispatcher;
};


/**
 * A PacketDispatcher opens an NFQ connection and will read packets
 * from the connection and cause the packets to be dispatched on
 * PacketQueues.  One (and only one) PacketDispatcher needs to be
 * instantiated and run in order to receive packets on PacketQueues.
 */
class PacketDispatcher
{
public:
    /**
     * Constructor: opens a connection to the userspace netfilter packet
     * queuing library.  One or more PacketQueue objects should be created
     * in order to receive packets.  After the PacketQueue objects are
     * created, run() should be called to begin dispatching packets.
     *
     * @param maxPacketSize Specify maximum amount of data to be
     *                      copied to userspace.
     * @throws PacketQueueError if something failed initializing NFQ
     */
    PacketDispatcher(int maxPacketSize) throw(PacketQueueError);

    virtual ~PacketDispatcher() throw();

    /**
     * Add a PacketQueue to the dispatcher.  This should be called
     * before run() is called.  The PacketQueue will automatically
     * be deleted when the dispatcher is stopped.  Thus, generally,
     * once a PacketQueue is created and passed to this method, the
     * creater of the PacketQueue would do best to not reference it
     * any longer since it may be deleted at any time.
     *
     * @param queue PacketQueue to add to this dispatcher
     * @throws PacketQueueError if there was an error creating the NPQ queue
     */
    void addQueue(PacketQueue *queue) throw(PacketQueueError);

    /**
     * Run the PacketDispatcher.  This method will not exit unless there
     * is an NPQ error, there are no more PacketQueues, or stop() is called.
     * Thus, PacketQueues should be added with addQueue() before this
     * method is called.
     *
     * When this method exits, the PacketDispatcher will be closed,
     * closing all connections to the NPQ subsystem and deleting all of
     * the PacketQueues.
     */
    void run() throw();

    /**
     * Stop the PacketDispatcher.  This will cause run() to exit.
     * This will block until the dispatcher has closed completely.
     */
    void stop() throw();

    /**
     * Returns true if this PacketDispatcher is closed and no longer
     * functional.. can be destroyed.
     */
    bool isClosed() throw();


private:
    friend class PacketQueue;

    /// Close connections to the NPQ subsystem and delete all PacketQueues
    void close() throw();

    /// Remove queue from queue set -- only done internally upon error.
    void removeQueue(PacketQueue *queue) throw();

    struct nfq_handle *handle() throw();

    set<PacketQueue *> m_queueSet;
    Mutex m_queueSetMutex;
    Mutex m_nfqMutex;
    ConditionVar m_closeCondition;

    bool m_stop;
    bool m_closed;
    struct nfq_handle *m_nfqHandle;
    const int m_bufsiz;
    unsigned char *m_buf;
    int m_nfqFD;
};


#endif // __PacketQueue_H__
