//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: Socket
//     Author: Jim Guggemos
//     Description: Implementation of the Socket class
//
// $Date$
//
//***************************************************************************

#include "Socket.h"
#include <iostream>
#include <string.h>

// Special throw for our DataExceptions
#define DEthrow(a,b,data)  throw( a(b,__FILE__, __PRETTY_FUNCTION__,__LINE__,(data)) )

#ifndef lthrow
#define lthrow(a,b) throw( a(b,__FILE__, __PRETTY_FUNCTION__,__LINE__) )
#endif


// Define macros to acquire Mutex::Holder object(s).  This must be
// done as macros (and not methods) since the scope is important.

// Lock the recv mutex
#define SOCKET_RECV_LOCK() Mutex::Holder _recvlock( this->m_recvMutex )

// Lock the send mutex
#define SOCKET_SEND_LOCK() Mutex::Holder _sendlock( this->m_sendMutex )

// Lock both mutex; used for making control changes to the socket (eg,
// calling ::bind() or ::connect()).
#define SOCKET_CONTROL_LOCK() SOCKET_RECV_LOCK(); SOCKET_SEND_LOCK()


// Static class members
set<Socket *> Socket::m_socketSet;
Mutex Socket::m_socketSetMutex( "Socket::m_socketSetMutex");


/// Opens an INET/STREAM (tcp) socket.
Socket::Socket() __SOCKTHROW : m_fd( -1 ), m_recvMutex( "Socket::recvMutex" ),
			       m_sendMutex( "Socket::sendMutex" )
{
    if ( (m_fd = socket( AF_INET, SOCK_STREAM, 0 )) < 0 )
	lthrow( Socket::SysCallException, "Socket::Socket()" );

    { // Make socket reusable in case we restart.
        int reuseaddr = 1;
        setsockopt( m_fd, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
                    sizeof(int) );
    }


    Mutex::Holder _lock(m_socketSetMutex);
    m_socketSet.insert( this );
}

/// Allocates a Socket with an already opened socket
Socket::Socket(int fd) __SOCKTHROW : m_fd(fd),
				     m_recvMutex( "Socket::recvMutex" ),
				     m_sendMutex( "Socket::sendMutex" )
{
    Mutex::Holder _lock(m_socketSetMutex);
    m_socketSet.insert( this );
}

/// Copy constructor
Socket::Socket(const Socket& s) throw() : m_fd(s.m_fd),
				      m_recvMutex( "Socket::recvMutex" ),
				      m_sendMutex( "Socket::sendMutex" ),
				      m_recvBuffer(s.m_recvBuffer)
{
    Mutex::Holder _lock(m_socketSetMutex);
    m_socketSet.insert( this );
}

/// Upon destruction, ensure the socket is properly closed.
Socket::~Socket()
{
    closeImmediate();

    {
	Mutex::Holder _lock(m_socketSetMutex);
	m_socketSet.erase( this );
    }
}

/// Copy assignment
Socket &Socket::operator=(const Socket &s) throw()
{
	m_fd = s.m_fd;
	m_recvBuffer = s.m_recvBuffer;
	return *this;
}

/// Binds socket to port, either using ANY or LOCALHOST as the address.
void
Socket::bind( SpecialAddr addr, int port ) __SOCKTHROW
{
    struct sockaddr_in in_addr;
    in_addr.sin_family = AF_INET;
    in_addr.sin_port = htons( port );
    in_addr.sin_addr.s_addr = htonl( (addr == ADDR_LOCALHOST) ?
				     INADDR_LOOPBACK : INADDR_ANY );
    bzero( &(in_addr.sin_zero), 8 );

    /// LOCK both socket mutexes
    SOCKET_CONTROL_LOCK();

    if ( ::bind( m_fd, (struct sockaddr*)&in_addr,
		 sizeof(struct sockaddr_in) ) < 0 )
	lthrow( Socket::SysCallException, "Socket::bind()" );
}

/** Returns port number to which this socket is bound; calling on an
    unbound socket will result in an exception. */
unsigned int
Socket::getPortNum() __SOCKTHROW
{
    struct sockaddr_in name;
    socklen_t len = sizeof( name );

    /// LOCK both socket mutexes
    SOCKET_CONTROL_LOCK();

    if ( ::getsockname( m_fd, (struct sockaddr*)&name, &len ) < 0 )
	lthrow( Socket::SysCallException, "Socket::getPortNum()" );

    return ntohs( name.sin_port );
}

/// Set Socket to listen with the given backlog queue length.
void
Socket::listen( int backlog ) __SOCKTHROW
{
    /// LOCK both socket mutexes
    SOCKET_CONTROL_LOCK();

    if ( ::listen( m_fd, backlog ) < 0 )
	lthrow( Socket::SysCallException, "Sock::listen()" );
}

/// Connect to addr:port; returns true if successful, false otherwise.
bool
Socket::connect( SpecialAddr addr, int port ) __SOCKTHROW
{
    struct in_addr in_addr;
    in_addr.s_addr = (addr == ADDR_LOCALHOST) ? INADDR_LOOPBACK : INADDR_ANY;

    return connect( in_addr, port );
}

/// Connect to addr:port; returns true if successful, false otherwise.
/// Note that addr is in HOST order, not network order.
bool
Socket::connect( const struct in_addr &addr, int port ) __SOCKTHROW
{
    struct sockaddr_in in_addr;
    in_addr.sin_family = AF_INET;
    in_addr.sin_port = htons( port );
    in_addr.sin_addr.s_addr = htonl( addr.s_addr );
    bzero( &(in_addr.sin_zero), 8 );

    /// LOCK both socket mutexes
    SOCKET_CONTROL_LOCK();

    if ( ::connect( m_fd, (struct sockaddr *)&in_addr,
		  sizeof(struct sockaddr_in) ) < 0 )
    {
	if ( errno == ECONNREFUSED )
	    return false;

	lthrow( Socket::SysCallException, "Sock::connect()" );
    }

    return true;
}

/** Wait for a connection and return a NEWLY ALLOCATED Socket object.
    Caller is responsible to free the returned Socket.  Returned
    Socket pointer will never be NULL - throws exception on error. */
Socket *
Socket::accept() __SOCKTHROW
{
    int new_fd;
    struct sockaddr_in their_addr;
    socklen_t size = sizeof(struct sockaddr_in);

    /// LOCK both socket mutexes
    SOCKET_CONTROL_LOCK();

    if ( (new_fd = ::accept( m_fd, (struct sockaddr*)&their_addr,
			     &size )) < 0 )
	lthrow( Socket::SysCallException, "Sock::accept()" );

    return new Socket( new_fd );
}

/// Sends the string over the socket; returns number of chars sent.
int
Socket::send( const string &buffer ) __SOCKTHROW
{
    int rval;

    // LOCK the send mutex
    SOCKET_SEND_LOCK();

    if ( (rval = ::send( m_fd, (void*)buffer.c_str(), buffer.size(), 0 )) < 0 )
    {
	if ( errno == ECONNRESET || errno == EPIPE )
	    lthrow( Socket::ClosedException, "Socket closed" );

	lthrow( Socket::SysCallException, "Sock::send()" );
    }

    return rval;
}

/** Receives available data from the socket.  Will block if there is no
    available data; however, if any data is available, it will be returned
    immediately regardless of length.  At most, only max_size bytes will
    be read.
*/
string
Socket::recv( int max_size ) __SOCKTHROW
{
    // LOCK the receive mutex
    SOCKET_RECV_LOCK();

    return _recv( max_size );
}

/// Internal version of recv() that DOES NOT ACQUIRE recvMutex.
string
Socket::_recv( int max_size ) __SOCKTHROW
{
    // Cap the maximum size of the buffer for ::recv().
    max_size = min( max_size, 16384 );
    char tbuf[max_size];

    int rval;

    // There was data left in the recv buffer by recv_line(), use that
    // first!
    if ( !m_recvBuffer.empty() )
    {
	int rbs = m_recvBuffer.size();
	rval = min( max_size, rbs );
	strncpy( tbuf, m_recvBuffer.c_str(), rval );

	m_recvBuffer.erase( 0, rval );
    }

    // Read from the socket
    else
    {
	if ( (rval = ::recv( m_fd, tbuf, max_size, 0 )) < 0 )
	    lthrow( Socket::SysCallException, "Sock::recv()" );
	
	if ( rval == 0 )
	    lthrow( Socket::ClosedException, "Socket closed" );

	string x( tbuf, rval );
    }

    string buffer( tbuf, min( max_size, rval ) );
    return buffer;
}

/** Receives data on the socket until the terminator is reached or
    until max_size is reached.  If max_size is reached without a
    terminator, Socket::LineTooLongException is thrown.
    Throws Socket::IncompleteLastLineException if socket was closed with
    unterminated data (data can be retrieved from exception).
*/
string
Socket::recv_line( int max_size, char terminator ) __SOCKTHROW
{
    string line;
    // Reserve space in the return string in advance for optimization sake;
    // however, cap it at a reasonable size.
    line.reserve( min(max_size,4096) );
    int len_so_far = 0;

    // LOCK the receive mutex
    SOCKET_RECV_LOCK();

    for ( ;; )
    {
	if ( len_so_far >= max_size )
	    DEthrow( Socket::LineTooLongException,
		     "Socket::recv_line(): Line too long before terminator",
		     line );

	string buf;
	try
	{
	    buf = _recv( max_size - len_so_far );
	}
	catch ( Socket::ClosedException )
	{
	    // If the socket is closed but we have unterminated data,
	    // throw the exception.
	    if ( ! line.empty() )
		DEthrow( Socket::IncompleteLastLineException,
		 "Socket::recv_line(): Socket closed with incomplete data.",
			 line );

	    // Otherwise, pass up the exception
	    throw;
	}

	unsigned int term_idx;
	if ( (term_idx = buf.find( terminator )) != string::npos )
	{
	    string useful;

	    // We want to INCLUDE the terminator
	    useful = buf.substr( 0, term_idx+1 );

	    // Anything else we stuff in the overflow buffer which _recv
	    // before actually calling ::recv.
	    try
	    {
		m_recvBuffer = buf.substr( term_idx + 1 );
	    }
	    catch ( std::out_of_range )
	    {
		m_recvBuffer = "";
	    }

	    line += useful;
	    break;
	}

	else
	{
	    line += buf;
	    len_so_far += buf.size();
	}
    }

    return line;
}

/** Receives data on the socket until the terminator string is
    reached or until max_size is reached.  If max_size is reached
    without a complete terminator, Socket::LineTooLongException is
    thrown.  There may be data after the terminator that is not 
    returned.  It can be retreived by calling recv() or recv_line() again.
    Throws Socket::IncompleteLastLineException if socket was closed with
    unterminated data (data can be retrieved from exception).
*/
string
Socket::recv_line( int max_size, string terminator ) __SOCKTHROW
{
    string line;
    // Reserve space in the return string in advance for optimization sake;
    // however, cap it at a reasonable size.
    line.reserve( min(max_size,4096) );
    int len_so_far = 0;
    
    if ( terminator.empty() )
	lthrow( Socket::Exception,
		"Socket::recv_line(): terminator string is empty" );

    int termlen = terminator.length();

    // LOCK the receive mutex
    SOCKET_RECV_LOCK();

    for ( ;; )
    {
	if ( len_so_far >= max_size )
	    DEthrow( Socket::LineTooLongException,
		     "Socket::recv_line(): Line too long before terminator",
		     line );

	string buf;
	try
	{
	    buf = _recv( max_size - len_so_far );
	}
	catch ( Socket::ClosedException )
	{
	    // If the socket is closed but we have unterminated data,
	    // throw the exception.
	    if ( ! line.empty() )
		DEthrow( Socket::IncompleteLastLineException,
		 "Socket::recv_line(): Socket closed with incomplete data.",
			 line );

	    // Otherwise, pass up the exception
	    throw;
	}

	line += buf;

	unsigned int term_idx;
	if ( (term_idx = line.find( terminator, ((len_so_far <= termlen) ? 0 :
				     len_so_far - termlen) )) != string::npos )
	{
	    // Anything else we stuff in the overflow buffer which _recv
	    // before actually calling ::recv.
	    try
	    {
		m_recvBuffer = line.substr( term_idx + termlen );
	    }
	    catch ( out_of_range )
	    {
		m_recvBuffer = "";
	    }

	    line.erase( term_idx + termlen );
	    break;
	}
	len_so_far += buf.size();
    }

    return line;
}


/** Close the socket; does not throw an error if already closed.  If
    socket is in use, will block until mutexes are released. */
void
Socket::close() throw()
{
    /// LOCK both socket mutexes
    SOCKET_CONTROL_LOCK();

    closeImmediate();
}


/** Close the socket immediately; will not block.  This will cause
    anyone using the socket to throw a "bad file descriptor" (EBADF)
    SysCallException. */
void
Socket::closeImmediate() throw()
{
    if ( isOpen() )
    {
	::shutdown( m_fd, SHUT_RDWR );
	::close( m_fd );
    }

    m_fd = -1;
}


/** Close all open sockets now; note that this does not delete the Socket
    objects but does shutdown() and close() the sockets.  Should only
    be called when the program terminates. */
void
Socket::closeAllImmediate() throw()
{
    Mutex::Holder _lock(m_socketSetMutex);

    for ( set<Socket *>::iterator s = m_socketSet.begin();
	  s != m_socketSet.end(); ++s )
	(*s)->closeImmediate();
}


/// Inner class - exception that captures errno.
Socket::SysCallException::
SysCallException(const string &msg, const string &file,
		 const string &function, const int line) : m_errno( errno )
{
    if ( m_errno < 1 )
	init( msg, file, function, line );
    else
    {
	string new_msg( msg );
	char buf[1024];
	const char *err_buf;

	// This "GNU" form of strerror_r may use the passed buffer
	// and it may not; either way it returns a pointer to
	// the string and is re-entrant (hopefully!).
	err_buf = strerror_r( m_errno, buf, 1024 );

	new_msg += ": ";
	new_msg += err_buf;

	init( new_msg, file, function, line );
    }
}
