//***************************************************************************
//
// $Source$
// $Revision: 70152 $
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: Socket
//     Author: Jim Guggemos
//     Description: Definition for the Socket class
//
// $Date: 2010-06-07 11:00:31 -0600 (Mon, 07 Jun 2010) $
//
//***************************************************************************

#ifndef __Socket_H__
#define __Socket_H__

#include "Mutex.h"
#include <string>
#include <sys/socket.h>
#include <netinet/in.h>
#include <set>
#include <stdexcept>
#include <stdio.h>

using namespace std;

#define __SOCKTHROW throw( Socket::Exception )


/** Provides a thread-safe TCP socket object that can be written and read
    by multiple threads simultaneously.  Also provides higher level
    'recv' functions to read until end of line or until a string terminator
    (think XML).
*/

class Socket
{
public:
    ///// -- BEGIN SOCKET EXCEPTIONS INTERFACE

    // Base class for all Socket exceptions
    class Exception : public std::exception
    {
    public:
	Exception(const string &msg, const string &file,
		  const string &function, const int line) throw() : m_line(-1)
        { init(msg, file, function, line); }
        virtual ~Exception() throw() {}

        virtual const char *what() const throw() { return m_what.c_str(); }
        string getMsg() const { return m_msg; }

    protected:
	Exception() throw() : m_line(-1) {}

        void init(const string &msg, const string &file,
                  const string &function, const int line) throw()
        {
            m_msg = msg;
            m_line = line;
            m_function = function;
            m_file = file;

            char linestr[20];
            sprintf( linestr, "%d", m_line );
            m_what = m_file + ":";
            m_what += linestr;
            m_what += ":" + m_function + ": " + m_msg;
        }

    private:
        string m_msg;
        int m_line;
        string m_function;
        string m_file;
        string m_what;
    };

    // Exception that stores errno and attaches it to the msg.
    class SysCallException : public Exception
    {
    private:
	const int m_errno;
    
    public:
	SysCallException(const string &msg, const string &file,
			 const string &function, const int line);
	int getErrno() { return m_errno; }
    };

    // Exception on read/write to a closed socket.
    class ClosedException : public Exception
    {
    public:
	ClosedException(const string &msg, const string &file,
			const string &function, const int line) :
            Exception( msg, file, function, line ) {}
    };

    // Base class that stores a line of data from the socket as well
    class DataException : public Exception
    {
    protected:
	string m_data;
    public:
	DataException(const string &msg, const string &file,
		      const string &function, const int line,
		      const string &data) :
	    Exception( msg, file, function, line ), m_data(data) {}
	virtual ~DataException() throw() {}
	const string &data() { return m_data; }
    };

    // Use data() to retrieve the offending line
    class LineTooLongException : public DataException
    {
    public:
	LineTooLongException(const string &msg, const string &file,
			     const string &function, const int line,
			     const string &data) :
	    DataException( msg, file, function, line, data ) {}
    };

    // Use data() to retrieve the last line that was not terminated properly
    class IncompleteLastLineException : public DataException
    {
    public:
	IncompleteLastLineException(const string &msg, const string &file,
				    const string &function, const int line,
				    const string &data) :
	    DataException( msg, file, function, line, data ) {}
    };


    
    ///// -- BEGIN SOCKET INTERFACE

    /// Opens an INET/STREAM (tcp) socket.
    Socket() __SOCKTHROW;

    /// Copy constructor
    Socket(const Socket& s) throw();

    /// Copy assignment
    Socket &operator=(const Socket &s) throw();

    /// Upon destruction, ensure the socket is properly closed.
    ~Socket();

    /// Returns true if socket is still open.
    bool isOpen() { return ( m_fd > -1 ); }

    enum SpecialAddr { ADDR_LOCALHOST, ADDR_ANY };

    /// Binds socket to port, either using ANY or LOCALHOST as the address.
    void bind( SpecialAddr addr, int port ) __SOCKTHROW;

    /** Returns port number to which this socket is bound; calling on an
	unbound socket will result in an exception. */
    unsigned int getPortNum() __SOCKTHROW;

    /// Set Socket to listen with the given backlog queue length.
    void listen( int backlog ) __SOCKTHROW;

    /// Connect to addr:port; returns true if successful, false otherwise.
    bool connect( SpecialAddr addr, int port ) __SOCKTHROW;

    /** Connect to addr:port; returns true if successful, false otherwise.
	Note that addr is in HOST order, not network order. */
    bool connect( const struct in_addr &addr, int port ) __SOCKTHROW;

    /** Wait for a connection and return a NEWLY ALLOCATED Socket object.
	Caller is responsible to free the returned Socket.  Returned
	Socket pointer will never be NULL - throws exception on error. */
    Socket *accept() __SOCKTHROW;

    /// Sends the string over the socket; returns number of chars sent.
    int send( const string &buffer ) __SOCKTHROW;

    /** Receives available data from the socket.  Will block if there is no
	available data; however, if any data is available, it will be returned
	immediately regardless of length.  At most, only max_size bytes will
	be read. */
    string recv( int max_size ) __SOCKTHROW;

    /** Receives data on the socket until the terminator is reached or
	until max_size is reached.  If max_size is reached without a
	terminator, Socket::LineTooLongException is thrown.
	Throws Socket::IncompleteLastLineException if socket was closed with
	unterminated data (data can be retrieved from exception). */
    string recv_line( int max_size, char terminator = '\n' ) __SOCKTHROW;

    /** Receives data on the socket until the terminator string is
	reached or until max_size is reached.  If max_size is reached
	without a complete terminator, Socket::LineTooLongException is
	thrown.  There may be data after the terminator that is not 
	returned.  It can be retreived by calling recv() or recv_line() again.
	Throws Socket::IncompleteLastLineException if socket was closed with
	unterminated data (data can be retrieved from exception). */
    string recv_line( int max_size, string terminator ) __SOCKTHROW;

    /** Close the socket; does not throw an error if already closed.  If
	socket is in use, will block until mutexes are released. */
    void close() throw();

    /** Close the socket immediately; will not block.  This will cause
	anyone using the socket to throw a "bad file descriptor" (EBADF)
	SysCallException. */
    void closeImmediate() throw();


    /** Close all open sockets now; note that this does not delete the Socket
	objects but does shutdown() and close() the sockets.  Should only
	be called when the program terminates. */
    static void closeAllImmediate() throw();

protected:
    /// Allocates a Socket with an already opened socket
    Socket(int fd) __SOCKTHROW;

    /// Internal version of recv() that DOES NOT acquire recv mutex.
    string _recv( int max_size ) __SOCKTHROW;

private:
    int m_fd;
    Mutex m_recvMutex;
    Mutex m_sendMutex;
    string m_recvBuffer;

    static set<Socket *> m_socketSet;
    static Mutex m_socketSetMutex;
};

#endif // __Socket_H__
